#!/usr/bin/env python3

"""
A module used to handle binary ntSecurityDescriptor from Active Directory LDAP.
"""

from struct import unpack

from ldap3.protocol.formatters.formatters import format_sid, format_uuid

SDDLTypeFlags = {
	'Self Relative'					: 0b1000000000000000,
	'RM Control Valid'				: 0b0100000000000000,
	'SACL Protected'				: 0b0010000000000000,
	'DACL Protected'				: 0b0001000000000000,
	'SACL Auto Inherit'				: 0b0000100000000000,
	'DACL Auto Inherit'				: 0b0000010000000000,
	'SACL Auto Inherit Required'	: 0b0000001000000000,
	'DACL Auto Inherit Required'	: 0b0000000100000000,
	'Server Security'				: 0b0000000010000000,
	'DACL Trusted'					: 0b0000000001000000,
	'SACL Defaulted'				: 0b0000000000100000,
	'SACL Present'					: 0b0000000000010000,
	'DACL Defaulted'				: 0b0000000000001000,
	'DACL Present'					: 0b0000000000000100,
	'Group Defaulted'				: 0b0000000000000010,
	'Owner Defaulted'				: 0b0000000000000001,
}

SID_SIZE = 28


def parse_ntSecurityDescriptor(input_buffer):
	""" Parses a ntSecurityDescriptor.
	"""
	out = dict()
	fields = ('Revision', 'Raw Type', 'Offset to owner SID', 'Offset to group SID', 'Offset to SACL', 'Offset to DACL')

	for k, v in zip(fields, unpack('<HHIIII', input_buffer[:20])):
		out[k] = v

	out['Type'] = parse_sddl_type(out['Raw Type'])

	for x in ('Owner', 'Group'):
		offset = out['Offset to %s SID' % (x.lower())]
		out['%s SID' % x] = format_sid(input_buffer[offset:offset + SID_SIZE])

	if out['Type']['SACL Present']:
		out['SACL'] = parse_acl(input_buffer[out['Offset to SACL']:])

	if out['Type']['DACL Present']:
		out['DACL'] = parse_acl(input_buffer[out['Offset to DACL']:])

	return out


def resolve_flags(bfr, flags):
	"""
	Helper to resolve flag values and names.

	Arguments:
		#bfr:integer
			The buffer containing flags.
		#flags:dict
			The dictionary of flag names and values.
	"""
	return {k: v & bfr != 0 for k, v in flags.items()}


def parse_sddl_type(typeflags):
	"""
	Parses SDDL Type flags.
	"""
	return resolve_flags(typeflags, SDDLTypeFlags)


def parse_acl(input_buffer):
	"""
	Parses ACL from SDDL.

	Returns a list of ACEs.
	"""
	out = dict()
	fields = ('Revision', 'Size', 'Num ACEs')

	for k, v in zip(fields, unpack('<HHI', input_buffer[:8])):
		out[k] = v

	out['ACEs'] = parse_aces(input_buffer[8:8 + out['Size']], out['Num ACEs'])
	return out


def parse_aces(input_buffer, count):
	"""
	Parses the list of ACEs.
	"""
	out = []
	while len(out) < count:
		ace = dict()
		fields = ('Raw Type', 'Raw Flags', 'Size', 'Raw Access Required')
		for k, v in zip(fields, unpack('<BBHI', input_buffer[:8])):
			ace[k] = v

		ace['Type'] = parse_sddl_dacl_ace_type(ace['Raw Type'])

		ace['Access Required'] = parse_ace_access(ace['Raw Access Required'])

		offset = 8

		if ace['Type'].endswith('Object'):
			fields = ('Raw Object Flags', )
			for k, v in zip(fields, unpack('<I', input_buffer[8:12])):
				ace[k] = v
			ace['Object Flags'] = parse_ace_object_flags(ace['Raw Object Flags'])

			offset = 12
			if ace['Object Flags']['Object Type Present']:
				ace['GUID'] = format_uuid(input_buffer[offset:offset + 16])
				offset += 16
			if ace['Object Flags']['Inherited Object Type Present']:
				ace['Inherited GUID'] = format_uuid(input_buffer[offset:offset + 16])
				offset += 16

			ace['SID'] = format_sid(input_buffer[offset:ace['Size']])

		ace['SID'] = format_sid(input_buffer[offset:ace['Size']])

		input_buffer = input_buffer[ace['Size']:]

		out.append(ace)
	return out


ACEAccessFlags = {
	'Generic Read'			: 0b10000000000000000000000000000000,
	'Generic Write'			: 0b01000000000000000000000000000000,
	'Generic Execute'		: 0b00100000000000000000000000000000,
	'Generic All'			: 0b00010000000000000000000000000000,
	'Maximum Allowed'		: 0b00000010000000000000000000000000,
	'Access SACL'			: 0b00000000100000000000000000000000,
	'Synchronise'			: 0b00000000000100000000000000000000,
	'Write Owner'			: 0b00000000000010000000000000000000,
	'Write DAC'				: 0b00000000000001000000000000000000,
	'Read Control'			: 0b00000000000000100000000000000000,
	'Delete'				: 0b00000000000000010000000000000000,
	'Ads Control Access'	: 0b00000000000000000000000100000000,
	'Ads List Object'		: 0b00000000000000000000000010000000,
	'Ads Delete Tree'		: 0b00000000000000000000000001000000,
	'Ads Write Prop'		: 0b00000000000000000000000000100000,
	'Ads Read Prop'			: 0b00000000000000000000000000010000,
	'Ads Self Write'		: 0b00000000000000000000000000001000,
	'Ads List'				: 0b00000000000000000000000000000100,
	'Ads Delete Child'		: 0b00000000000000000000000000000010,
	'Ads Create Child'		: 0b00000000000000000000000000000001
}


def parse_ace_access(input_buffer):
	"""
	Parses access flags in an ACE.
	"""
	return resolve_flags(input_buffer, ACEAccessFlags)


ACEObjectFlags = {
	'Object Type Present'				: 0b00000000000000000000000000000010,
	'Inherited Object Type Present'		: 0b00000000000000000000000000000001
}


def parse_ace_object_flags(input_buffer):
	"""
	Parses flags in an ACE containing an object.
	"""
	return resolve_flags(input_buffer, ACEObjectFlags)


ACEType = {
	0x00: 'Access Allowed',
	0x01: 'Access Denied',
	0x02: 'System Audit',
	0x03: 'System Alarm',
	0x04: 'Access Allowed Compound',
	0x05: 'Access Allowed Object',
	0x06: 'Access Denied Object',
	0x07: 'System Audit Object',
	0x08: 'System Alarm Object',
	0x09: 'Access Allowed Callback',
	0x0A: 'Access Denied Callback',
	0x0B: 'Access Allowed Callback Object',
	0x0C: 'Access Denied Callback Object',
	0x0D: 'System Audit Callback',
	0x0E: 'System Alarm Callback',
	0x0F: 'System Audit Callback Object',
	0x10: 'System Alarm Callback Object',
	0x11: 'System Mandatory Label',
	0x12: 'System Resource Attribute',
	0x13: 'System Scoped Policy ID'
}


def parse_sddl_dacl_ace_type(ace_type):
	"""
	Parses the type of an ACE.
	"""
	return ACEType[ace_type]




from sys import __stdout__, __stderr__, exit
from termcolor import colored
from multiprocessing.dummy import Pool as ThreadPool
from tqdm import tqdm

import dns.resolver


def info(content):
	__stderr__.write("%s\n" % colored("[+] " + str(content), "blue", attrs=["bold"]))


def error(content):
	__stderr__.write("%s\n" % colored("[!] " + str(content), "red", attrs=["bold"]))
	exit(1)


class Logger(object):

	def __init__(self, outfile=None, quiet=False):
		self.quiet = quiet
		self.terminal = __stdout__
		self.log = open(outfile, 'w') if outfile else None

	def write(self, message):
		if not self.quiet:
			self.terminal.write(message)
		if self.log:
			self.log.write(message)

	def flush(self):
		if self.log:
			self.log.flush()
		pass


class ResolverThread(object):

	def __init__(self, dns_server):
		self.dns_server = dns_server
		self.resolutions = []

	def resolve(self, hostname):
		if self.dns_server:
			resolver = dns.resolver.Resolver()
			resolver.nameservers = [self.dns_server]
		else:
			resolver = dns.resolver
		try:
			answers = resolver.query(hostname, 'A', tcp=True)
			for rdata in answers:
				if rdata.address:
					self.resolutions.append({
						"hostname": hostname,
						"address": rdata.address
					})
					break
			else:
				pass
		except Exception:
			pass


def resolve(hostnames, dns_server):
	pool = ThreadPool(20)
	resolver_thread = ResolverThread(dns_server)
	with tqdm(total=len(hostnames)) as pbar:
		for _ in pool.imap_unordered(resolver_thread.resolve, tqdm(hostnames, desc="Resolution", bar_format="{desc} {n_fmt}/{total_fmt} hostnames")):
			pbar.update()
	pool.close()
	pool.join()
	results_set = []
	for computer in resolver_thread.resolutions:
		results_set.append(computer)
	return results_set




from ldap3 import ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES
from ldap3.protocol.formatters.validators import validate_sid, validate_guid

ALL_ATTRIBUTES = ALL_ATTRIBUTES
ALL_OPERATIONAL_ATTRIBUTES = ALL_OPERATIONAL_ATTRIBUTES
ALL = [ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES]
validate_sid = validate_sid
validate_guid = validate_guid


class ActiveDirectoryView(object):
	"""
	Manage a view of a Active Directory.
	"""

	class ActiveDirectoryInvalidSID(Exception):
		pass

	class ActiveDirectoryInvalidGUID(Exception):
		pass



from os import path
from json import load as json_load
from ldeep.views.activedirectory import ActiveDirectoryView, ALL, ALL_ATTRIBUTES, validate_guid, validate_sid
from ldeep.views.constants import WELL_KNOWN_SIDS


class UnexpectedFormatException(Exception):
    pass


# case insenitive equal when x and y are str instances
def eq(x, y):
    if isinstance(x, str) and isinstance(y, str):
        return x.lower() == y.lower()
    else:
        return x == y

# respect the ANR field
def eq_anr(record, value):

    def fmap(f, obj):
        if isinstance(obj, dict):
            return any(fmap(f, sub) for sub in obj.values())
        elif isinstance(obj, list):
            return any(fmap(f, k) for k in obj)
        elif isinstance(obj, str):
            return f(obj)
        else:
            raise UnexpectedFormatException(f"Unexpected value, expected: dict, list or str, obtained: {type(obj)}.")

    validate = lambda x: x.lower().startswith(value.lower())

    keys = ["displayName", "givenName", "legacyExchangeDN", "physicalDeliveryOfficeName", "proxyAddresses", "Name", "sAMAccountName", "sn"]
    for k in keys:
        if k in record and fmap(validate, record[k]):
            return True

class CacheActiveDirectoryView(ActiveDirectoryView):

    # Constant functions (first arg -> self but we don't need it)
    USER_LOCKED_FILTER = lambda _: {"files": ["users_locked"]}
    GROUPS_FILTER = lambda _: {"files": ["groups"]}
    USER_ALL_FILTER = lambda _: {"files": ["users_all"]}
    USER_SPN_FILTER = lambda _: {"files": ["users_spn"]}
    COMPUTERS_FILTER = lambda _: {"files": ["machines"]}
    ANR = lambda _, u: {"files": ["users_all", "groups", "machines"], "filter": lambda record: eq_anr(record, u)}
    GROUP_DN_FILTER = lambda _, g: {"fmt": "json", "files": ["groups"], "filter": lambda x: eq(x["sAMAccountName"], g)}
    ACCOUNTS_IN_GROUP_FILTER = lambda _, p, g: {
        "fmt": "json",
        "files": ["users_all", "groups", "machines"],
        "filter": lambda x: ("primaryGroupID" in x and eq(p, x["primaryGroupID"])) or ("memberOf" in x and g in x["memberOf"])
    }
    ACCOUNT_IN_GROUPS_FILTER = lambda _, u: {
        "fmt": "json",
        "files": ["users_all", "groups", "machines"],
        "filter": lambda x: eq(x["sAMAccountName"], u)
    }
    DISTINGUISHED_NAME = lambda _, n: {
        "fmt": "json",
        "files": ["users_all", "groups", "machines"],
        "filter": lambda x: eq(x["distinguishedName"], n)
    }
    PRIMARY_GROUP_ID = lambda _, i: {
        "fmt": "json",
        "files": ["users_all", "groups", "machines"],
        "filter": lambda x: x["objectSid"].endswith(f"-{i}")
    }
    AUTH_POLICIES_FILTER = lambda _: {"files": ["auth_policies"]}
    SILOS_FILTER = lambda _: {"files": ["silos"]}
    SILO_FILTER = lambda _, s: {
        "fmt": "json",
        "files": ["silos"],
        "filter": lambda x: eq(x["cn"], s)
    }

    # Not implemented:
    DOMAIN_INFO_FILTER = lambda _: None
    GPO_INFO_FILTER = lambda _: None
    OU_FILTER = lambda _: None
    PSO_INFO_FILTER = lambda _: None
    TRUSTS_INFO_FILTER = lambda _: None
    ZONES_FILTER = lambda _: None
    ZONE_FILTER = lambda _: None
    USER_ACCOUNT_CONTROL_FILTER = lambda _, __: None
    USER_ACCOUNT_CONTROL_FILTER_NEG = lambda _, __: None
    USER_LOCKED_FILTER = lambda _: None
    SMSA_FILTER = lambda _: None
    SHADOW_PRINCIPALS_FILTER = lambda _: None
    UNCONSTRAINED_DELEGATION_FILTER = lambda _: None
    CONSTRAINED_DELEGATION_FILTER = lambda _: None
    RESOURCE_BASED_CONSTRAINED_DELEGATION_FILTER = lambda _: None
    ALL_DELEGATIONS_FILTER = lambda _: None

    class CacheActiveDirectoryException(Exception):
        pass

    class CacheActiveDirectoryDirNotFoundException(Exception):
        pass

    def __init__(self, cache_dir=".", prefix="ldeep_"):
        """
        CacheActiveDirectoryView constructor.
        Initialize the cache state with the provided directory and file prefixes.

        @cache_dir: directory containing ldeep files
        @prefix: prefix of the files
        """
        if not path.exists(cache_dir):
            raise self.CacheActiveDirectoryDirNotFoundException(f"{cache_dir} doesn't exist.")
        self.path = cache_dir
        self.prefix = prefix
        self.fqdn, self.base_dn = self.__get_domain_info()
        self.attributes = ALL

    def set_all_attributes(self, attributes=ALL):
        self.attributes = attributes

    def all_attributes(self):
        return self.attributes

    def set_controls(self, controls):
        pass

    def query(self, cachefilter, attributes=[], base=None, scope=None, **filter_args):
        """
        Perform a query to cache files.

        @cachefilter: a dict containing the following fields: fmt (optional), files and filter (optional).
        @attributes: only use to deduce the file formats to use (`lst` or `json`).
        @base: Not implemented.
        @scope: Not implemented.

        @return a list of records.
        """

        def scrub_json_from_key(obj, func):
            if isinstance(obj, dict):
                for key in list(obj.keys()):
                    if func(key):
                        del obj[key]
                    else:
                        scrub_json_from_key(obj[key], func)
            elif isinstance(obj, list):
                for k in reversed(range(len(obj))):
                    if func(obj[k]):
                        del obj[k]
                    else:
                        scrub_json_from_key(obj[k], func)

        # Process unimplemented queries
        if cachefilter is None:
            raise self.CacheActiveDirectoryException("Cache query not supported.")

        # Get format of cache files to use: either `lst` or `json`
        if "fmt" in cachefilter:
            fmt = cachefilter["fmt"]
        elif ALL_ATTRIBUTES in attributes:
            fmt = "json"
        else:
            fmt = "lst"

        data = []
        # For each file, retrieve result based on an optional filter
        for fil in cachefilter["files"]:
            filename = "{prefix}_{file}.{ext}".format(
                        prefix=self.prefix,
                        file=fil,
                        ext=fmt)

            with open(path.join(self.path, filename)) as fp:
                # Two cases
                # all attributes are required thus we parse the JSON file
                if fmt == "json":
                    json = json_load(fp)

                    if "ntSecurityDescriptor" not in self.attributes:
                        scrub_json_from_key(json, lambda x: x == "nTSecurityDescriptor")

                    if "filter" in cachefilter:
                        for record in json:
                            if cachefilter["filter"](record):
                                data.append(record)
                    else:
                        data += json

                # we use the lst file
                else:
                    if "filter" in cachefilter:
                        for line in fp:
                            x = {"sAMAccountName": line.strip()}  # a little hacky :)
                            if cachefilter["filter"](x):
                                data += [line.strip()]
                    else:
                        data += map(lambda x: x.strip(), fp.readlines())
        return data

    def resolve_sid(self, sid):
        """
        Two cases:
            * the SID is a WELL KNOWN SID and a local SID, the name of the corresponding account is returned;
            * else, the SID is search through the cache and the corresponding record is returned.

        @sid: the sid to search for.

        @throw ActiveDirectoryInvalidSID if the SID is not a valid SID.
        @return the record corresponding to the SID queried.
        """
        if sid in WELL_KNOWN_SIDS:
            return WELL_KNOWN_SIDS[sid]
        elif validate_sid(sid):
            results = self.query({
                "fmt": "json",
                "files": ["users_all", "groups", "machines"],
                "filter": lambda x: x["objectSid"] == sid
            })
            if results:
                return results
        raise self.ActiveDirectoryInvalidSID(f"SID: {sid}")

    def resolve_guid(self, guid):
        """
        Return the cache record with the provided GUID.

        @guid: the guid to search for.

        @throw ActiveDirectoryInvalidGUID if the GUID is not a valid GUID.
        @return the record corresponding to the guid queried.
        """
        if validate_guid(guid):
            results = self.query({
                "fmt": "json",
                "files": ["users_all", "groups", "machines"],
                "filter": lambda x: x["objectGUID"] == guid
            })
            if results:
                return results
        raise self.ActiveDirectoryInvalidGUID(f"GUID: {guid}")

    def get_sddl(self, *kwargs):
        raise NotImplementedError

    def __get_domain_info(self):
        """
        Private functions to retrieve the cache domain name.
        """
        filename = "{prefix}_domain_policy.lst".format(prefix=self.prefix)
        with open(path.join(self.path, filename)) as fp:
            for line in fp:
                if line.startswith("distinguishedName:"):
                    base = line.split(" ")[1].strip()
                    domain = base.replace("DC=", ".")[1:].replace(",", "")
                    return domain, base



#!/usr/bin/env python3

from argparse import ArgumentParser
from json import dump as json_dump
from base64 import b64encode
from math import fabs
from re import compile as re_compile
from datetime import date, datetime, timedelta
from commandparse import Command
from time import sleep

from Cryptodome.Cipher import AES
from Cryptodome.Hash import MD4, SHA1
from Cryptodome.Protocol.KDF import PBKDF2

from pyasn1.error import PyAsn1UnicodeDecodeError

from ldeep.views.activedirectory import ActiveDirectoryView, ALL, ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES
from ldeep.views.constants import USER_ACCOUNT_CONTROL, LDAP_SERVER_SD_FLAGS_OID_SEC_DESC
from ldeep.views.ldap_activedirectory import LdapActiveDirectoryView
from ldeep.views.cache_activedirectory import CacheActiveDirectoryView

from ldeep.utils import error, info, Logger, resolve as utils_resolve
from ldeep.utils.structure import Structure
from ldeep.utils.sddl import parse_ntSecurityDescriptor
from ldap3.protocol.formatters.formatters import format_sid

import sys


class Ldeep(Command):

    def __init__(self, query_engine, format="json"):
        self.engine = query_engine
        if format == "json":
            self.__display = self.__display_json

    def display(self, records, verbose=False, specify_group=True):
        def default(o):
            if isinstance(o, date) or isinstance(o, datetime):
                return o.isoformat()
            elif isinstance(o, bytes):
                return b64encode(o).decode('ascii')

        if verbose:
            self.__display(list(map(dict, records)), default)
        else:
            for record in records:
                if "objectClass" not in record:
                    print(record)
                elif "group" in record["objectClass"]:
                    print(record["sAMAccountName"] + (" (group)" if specify_group else ""))
                elif "user" in record["objectClass"]:
                    print(record["sAMAccountName"])
                elif "dnsNode" in record["objectClass"]:
                    print("{dc} {rec}".format(dc=record["dc"], rec=" ".join(record["dnsRecord"])))
                elif "dnsZone" in record["objectClass"]:
                    print(record["dc"])
                elif "domain" in record["objectClass"]:
                    print(record["dn"])
                elif "pKIEnrollmentService" in record["objectClass"]:
                    print(record["dNSHostName"])
                elif "msDS-AuthNPolicy" in record["objectClass"] or "msDS-AuthNPolicySilo" in record["objectClass"]:
                    print(record["cn"])

    def __display_json(self, records, default):
        json_dump(records, sys.stdout, ensure_ascii=False, default=default, sort_keys=True, indent=2)

    # LISTERS #

    def list_users(self, kwargs):
        """
        List users according to a filter.

        Arguments:
            @verbose:bool
                Results will contain full information
            @filter:string = ["all", "spn", "enabled", "disabled", "locked", "nopasswordexpire", "passwordexpired", "nokrbpreauth", "reversible"]
        """
        verbose = kwargs.get("verbose", False)
        filter_ = kwargs.get("filter", "all")

        if verbose:
            attributes = self.engine.all_attributes()
        else:
            attributes = ["samAccountName", "objectClass"]

        if filter_ == "all":
            results = self.engine.query(self.engine.USER_ALL_FILTER(), attributes)
        elif filter_ == "spn":
            results = self.engine.query(self.engine.USER_SPN_FILTER(), attributes)
        elif filter_ == "enabled":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER_NEG(USER_ACCOUNT_CONTROL["ACCOUNTDISABLE"]), attributes)
        elif filter_ == "disabled":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER(USER_ACCOUNT_CONTROL["ACCOUNTDISABLE"]), attributes)
        elif filter_ == "locked":
            results = self.engine.query(self.engine.USER_LOCKED_FILTER(), attributes)
        elif filter_ == "nopasswordexpire":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER(USER_ACCOUNT_CONTROL["DONT_EXPIRE_PASSWORD"]), attributes)
        elif filter_ == "passwordexpired":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER(USER_ACCOUNT_CONTROL["PASSWORD_EXPIRED"]), attributes)
        elif filter_ == "nokrbpreauth":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER(USER_ACCOUNT_CONTROL["DONT_REQ_PREAUTH"]), attributes)
        elif filter_ == "reversible":
            results = self.engine.query(self.engine.USER_ACCOUNT_CONTROL_FILTER(USER_ACCOUNT_CONTROL["ENCRYPTED_TEXT_PWD_ALLOWED"]), attributes)
        else:
            return None

        self.display(results, verbose)

    def list_groups(self, kwargs):
        """
        List the groups.

        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)

        if verbose:
            attributes = self.engine.all_attributes()
        else:
            attributes = ["samAccountName", "objectClass"]

        self.display(self.engine.query(self.engine.GROUPS_FILTER(), attributes), verbose, specify_group=False)

    def list_machines(self, kwargs):
        """
        List the machine accounts.

        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)

        if verbose:
            attributes = self.engine.all_attributes()
        else:
            attributes = ["samAccountName", "objectClass"]

        self.display(self.engine.query(self.engine.COMPUTERS_FILTER(), attributes), verbose, specify_group=False)

    def list_computers(self, kwargs):
        """
        List the computer hostnames and resolve them if --resolve is specify.

        Arguments:
            @resolve:bool
                A resolution on all computer names will be performed
            @dns:string
                An optional DNS server to use for the resolution
        """
        resolve = "resolve" in kwargs and kwargs["resolve"]
        dns = kwargs.get("dns", "")

        hostnames = []
        results = self.engine.query(self.engine.COMPUTERS_FILTER(), ["name"])
        for result in results:
            if "name" in result:  # ugly
                computer_name = result["name"]
            else:
                computer_name = result[:-1]  # removing trailing $ sign

            hostnames.append(f"{computer_name}.{self.engine.fqdn}")
            # print only if resolution was not mandated
            if not resolve:
                print(f"{computer_name}.{self.engine.fqdn}")
        # do the resolution
        if resolve:
            for computer in utils_resolve(hostnames, dns):
                print("{addr:20} {name}".format(addr=computer["address"], name=computer["hostname"]))

    def list_gmsa(self, kwargs):
        """
        List the gmsa accounts and retrieve NT hash if possible.
        """

        attributes = ["samAccountName", "objectClass", "msDS-ManagedPassword"]

        try:
            self.engine.ldap.start_tls()
        except Exception as e:
            print(f"Can't retrieve gmsa, TLS needed: {e}")
            return
        entries = self.engine.query("(ObjectClass=msDS-GroupManagedServiceAccount)", attributes)

        constants = b'\x6b\x65\x72\x62\x65\x72\x6f\x73\x7b\x9b\x5b\x2b\x93\x13\x2b\x93\x5c\x9b\xdc\xda\xd9\x5c\x98\x99\xc4\xca\xe4\xde\xe6\xd6\xca\xe4'
        iv = b'\x00' * 16

        for entry in entries:
            try:
                sam = entry['sAMAccountName']
                data = entry['msDS-ManagedPassword']
                blob = MSDS_MANAGEDPASSWORD_BLOB()
                blob.fromString(data)
                password = blob['CurrentPassword'][:-2]

                # Compute NT hash
                hash = MD4.new()
                hash.update(password)
                nthash = hash.hexdigest()

                # Quick and dirty way to get the FQDN of the account's domain
                dc_list = []
                for s in entry['dn'].split(','):
                    if s.startswith('DC='):
                        dc_list.append(s[3:])

                domain_fqdn = '.'.join(dc_list)
                salt = f'{domain_fqdn.upper()}host{sam[:-1].lower()}.{domain_fqdn.lower()}'
                encryption_key = PBKDF2(password.decode('utf-16-le', 'replace').encode(), salt.encode(), 32, count=4096, hmac_hash_module=SHA1)

                # Compute AES keys
                cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
                first_part = cipher.encrypt(constants)
                cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
                second_part = cipher.encrypt(first_part)
                aes256_key = first_part[:16] + second_part[:16]

                cipher = AES.new(encryption_key[:16], AES.MODE_CBC, iv)
                aes128_key = cipher.encrypt(constants[:16])

                print(f'{sam}:nt:{nthash}')
                print(f'{sam}:aes128-cts-hmac-sha1-96:{aes128_key.hex()}')
                print(f'{sam}:aes256-cts-hmac-sha1-96:{aes256_key.hex()}')
            except Exception as e:
                print(f"Raised exception: {e}")
                print(sam)

    def list_domain_policy(self, _):
        """
        Return the domain policy.
        """
        FILETIME_TIMESTAMP_FIELDS = {
            "lockOutObservationWindow": (60, "mins"),
            "lockoutDuration": (60, "mins"),
            "maxPwdAge": (86400, "days"),
            "minPwdAge": (86400, "days"),
            "forceLogoff": (60, "mins")
        }

        FOREST_LEVELS = {
            7: "Windows Server 2016",
            6: "Windows Server 2012 R2",
            5: "Windows Server 2012",
            4: "Windows Server 2008 R2",
            3: "Windows Server 2008",
            2: "Windows Server 2003",
            1: "Windows Server 2003 operating system through Windows Server 2016",
            0: "Windows 2000 Server operating system through Windows Server 2008 operating system"
        }

        FIELDS_TO_PRINT = [
            "dc",
            "distinguishedName",
            "lockOutObservationWindow",
            "lockoutDuration",
            "lockoutThreshold",
            "maxPwdAge",
            "minPwdAge",
            "minPwdLength",
            "pwdHistoryLength",
            "pwdProperties",
            "ms-DS-MachineAccountQuota",
            "msDS-Behavior-Version"]

        policy = self.engine.query(self.engine.DOMAIN_INFO_FILTER())
        if policy:
            policy = policy[0]
            for field in FIELDS_TO_PRINT:
                val = policy.get(field, None)
                if val is None:
                    continue

                if field == "lockOutObservationWindow" and isinstance(val, timedelta):
                    val = int(val.total_seconds()) / 60
                elif field in FILETIME_TIMESTAMP_FIELDS.keys() and type(val) == int:
                    val = int((fabs(float(val)) / 10**7) / FILETIME_TIMESTAMP_FIELDS[field][0])
                if field in FILETIME_TIMESTAMP_FIELDS.keys():
                    val = "%d %s" % (val, FILETIME_TIMESTAMP_FIELDS[field][1])
                if field == "msDS-Behavior-Version" and isinstance(val, int):
                    val = "%s" % (FOREST_LEVELS[policy[field]])

                print("%s: %s" % (field, val))

    def list_ou(self, _):
        """
        Return the list of organizational units with linked GPO.
        """
        cn_re = re_compile("{[^}]+}")
        results = self.engine.query(self.engine.GPO_INFO_FILTER(), ["cn", "displayName"])
        gpos = {}
        for gpo in results:
            gpos[gpo["cn"]] = gpo["displayName"]

        results = self.engine.query(self.engine.OU_FILTER())
        for result in results:
            print(result["distinguishedName"])
            if "gPLink" in result:
                guids = cn_re.findall(result["gPLink"])
                if len(guids) > 0:
                    print("[gPLink]")
                    print("* {}".format("\n* ".join([gpos[g] if g in gpos else g for g in guids])))

    def list_gpo(self, _):
        """
        Return the list of Group policy objects.
        """
        results = self.engine.query(self.engine.GPO_INFO_FILTER(), ["cn", "displayName"])
        for gpo in results:
            print("{cn}: {name}".format(cn=gpo["cn"], name=gpo["displayName"]))

    def list_pso(self, _):
        """
        List the Password Settings Objects.
        """
        FILETIME_TIMESTAMP_FIELDS = {
            "msDS-LockoutObservationWindow": (60, "mins"),
            "msDS-MinimumPasswordAge": (86400, "days"),
            "msDS-MaximumPasswordAge": (86400, "days"),
            "msDS-LockoutDuration": (60, "mins")
        }

        FIELDS_TO_PRINT = [
            "cn",
            "msDS-PasswordReversibleEncryptionEnabled",
            "msDS-PasswordSettingsPrecedence",
            "msDS-MinimumPasswordLength",
            "msDS-PasswordHistoryLength",
            "msDS-PasswordComplexityEnabled",
            "msDS-LockoutObservationWindow",
            "msDS-LockoutDuration",
            "msDS-LockoutThreshold",
            "msDS-MinimumPasswordAge",
            "msDS-MaximumPasswordAge",
            "msDS-PSOAppliesTo",
        ]

        psos = self.engine.query(self.engine.PSO_INFO_FILTER())
        for policy in psos:
            for field in FIELDS_TO_PRINT:
                val = policy.get(field, None)
                if not val:
                    continue
                if isinstance(val, list):
                    targets = []
                    for target in val:
                        targets.append(target)
                    val = " | ".join(targets)
                else:
                    val = policy[field]

                if field in FILETIME_TIMESTAMP_FIELDS.keys():
                    val = int((fabs(float(val)) / 10**7) / FILETIME_TIMESTAMP_FIELDS[field][0])
                    val = "{val} {typ}".format(val=val, typ=FILETIME_TIMESTAMP_FIELDS[field][1])
                print("{field}: {val}".format(field=field, val=val))

    def list_trusts(self, _):
        """
        List the domain's trust relationships.
        """
        results = self.engine.query(self.engine.TRUSTS_INFO_FILTER())
        FIELDS_TO_PRINT = [
            "dn",
            "cn",
            "securityIdentifier",
            "name",
            "trustDirection",
            "trustPartner",
            "trustType",
            "trustAttributes",
            "flatName"
        ]

        for result in results:
            for field in FIELDS_TO_PRINT:
                if field in result:
                    val = result[field]
                    if field == "trustDirection":
                        if int(val) == 0x00000003:
                            val = "bidirectional"
                        elif int(val) == 0x00000002:
                            val = "outbound"
                        elif int(val) == 0x00000001:
                            val = "inbound"
                        elif int(val) == 0x00000000:
                            val = "disabled"
                    elif field == "trustType":
                        if int(val) == 0x00000001:
                            val = "Non running Windows domain"
                        elif int(val) == 0x00000002:
                            val = "Windows domain running Active Directory"
                        elif int(val) == 0x00000003:
                            val = "Non Windows domain"
                    print("{field}: {val}".format(field=field, val=val))
            print("")

    def list_zones(self, kwargs):
        """
        List the DNS zones configured in the Active Directory.

        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)

        if not verbose:
            attributes = ["dc", "objectClass"]
        else:
            attributes = ALL

        self.display(
            self.engine.query(
                self.engine.ZONES_FILTER(),
                attributes, base=','.join(["CN=MicrosoftDNS,DC=DomainDNSZones", self.engine.base_dn])
            ),
            verbose
        )

    def list_pkis(self, kwargs):
        """
        List pkis.
        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)

        if verbose:
            attributes = self.engine.all_attributes()
        else:
            attributes = ["dNSHostName", "objectClass"]

        self.display(
            self.engine.query(
                self.engine.PKI_FILTER(),
                attributes,
                base=','.join(["CN=Enrollment Services,CN=Public Key Services,CN=Services,CN=CONFIGURATION", self.engine.base_dn]),
            ),
            verbose
        )

    # subnet_dn not used, TODO: remove?
    def list_subnets(self, kwargs):
        """
        List sites and associated subnets.
        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)

        if not verbose:
            attributes = ["distinguishedName", "name", "description"]
        else:
            attributes = ALL

        if verbose:
            self.display(
                self.engine.query(
                    self.engine.SITES_FILTER(),
                    attributes, base=','.join(["CN=Configuration", self.engine.base_dn])
                ),
                verbose
            )
        else:
            entries = self.engine.query(self.engine.SITES_FILTER(), attributes, base=','.join(["CN=Configuration", self.engine.base_dn]))

            site_dn = ""
            site_name = ""
            site_description = ""
            subnet_dn = ""
            subnet_name = ""
            subnet_description = ""
            for entry in entries:
                site_dn = entry["distinguishedName"] if entry["distinguishedName"] else ""
                site_name = entry["name"] if entry["name"] else ""
                site_description = entry["description"][0] if entry["description"] else ""
                subnet_entries = self.engine.query(self.engine.SUBNET_FILTER(site_dn), attributes, base=','.join(["CN=Sites,CN=Configuration", self.engine.base_dn]))
                for subnet in subnet_entries:
                    subnet_dn = subnet["distinguishedName"] if subnet["distinguishedName"] else ""
                    subnet_name = subnet["name"] if subnet["name"] else ""
                    subnet_description = subnet["description"][0] if subnet["description"] else ""

                output = "Site: {}".format(site_name)
                output += " | Subnet: {}".format(subnet_name) if subnet_name else ""
                output += " | Site description: {}".format(site_description) if site_description else ""
                output += " | Subnet description: {}".format(subnet_description) if subnet_description else ""
                print(output)

    def list_conf(self, kwargs):
        """
        Dump the configuration partition of the Active Directory.
        """
        self.display(
            self.engine.query(
                self.engine.ALL_FILTER(),
                ALL, base=','.join(["CN=Configuration", self.engine.base_dn])
            ),
            True
        )

    def list_auth_policies(self, kwargs):
        """
        List the authentication policies configured in the Active Directory.

        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)
        attributes = ALL if verbose else ["cn", "objectClass"]

        self.display(
            self.engine.query(
                self.engine.AUTH_POLICIES_FILTER(),
                attributes, base=','.join(["CN=AuthN Policy Configuration,CN=Services,CN=Configuration", self.engine.base_dn])
            ),
            verbose
        )

    def list_silos(self, kwargs):
        """
        List the silos configured in the Active Directory.

        Arguments:
            @verbose:bool
                Results will contain full information
        """
        verbose = kwargs.get("verbose", False)
        attributes = ALL if verbose else ["cn", "objectClass"]

        self.display(
            self.engine.query(
                self.engine.SILOS_FILTER(),
                attributes, base=','.join(["CN=AuthN Policy Configuration,CN=Services,CN=Configuration", self.engine.base_dn])
            ),
            verbose
        )

    def list_smsa(self, kwargs):
        """
        List the smsa accounts and the machines they are associated with.

        Arguments:
            @verbose:bool
                Results will contain full information
        """

        verbose = kwargs.get("verbose", False)
        attributes = ALL if verbose else ["samAccountName", "msDS-HostServiceAccountBL"]
        entries = self.engine.query(self.engine.SMSA_FILTER(), attributes)

        if verbose:
            self.display(entries, verbose)
        else:
            for entry in entries:
                sam = entry['sAMAccountName']
                for host in entry["msDS-HostServiceAccountBL"]:
                    print(f'{sam}:{host}')

    def list_shadow_principals(self, kwargs):
        """
        List the shadow principals and the groups associated with.

        Arguments:
            @verbose:bool
                Results will contain full information
        """

        verbose = kwargs.get("verbose", False)
        attributes = ALL if verbose else ["member", "msDS-ShadowPrincipalSid"]
        base=','.join(["CN=Shadow Principal Configuration,CN=Services,CN=Configuration", self.engine.base_dn])
        entries = self.engine.query(self.engine.SHADOW_PRINCIPALS_FILTER(), attributes, base=base)

        if verbose:
            self.display(entries, verbose)
        else:
            for entry in entries:
                print(f"User {entry['member'][0]} added to Group {format_sid(entry['msDS-ShadowPrincipalSid'])}")

    def list_delegations(self, kwargs):
        """
        List accounts configured for any kind of delegation.

        Arguments:
            @verbose:bool
                Results will contain full information
            @filter:string = ["all", "unconstrained", "constrained", "rbcd"]
        """
        verbose     = kwargs.get("verbose", False)
        filter_     = kwargs.get("filter", "all")

        attributes = ALL if verbose else ["samAccountName", "userAccountControl"]

        if filter_ == "all":
            if not verbose:
                attributes.extend([
                    "msDS-AllowedToDelegateTo",
                    "msDS-AllowedToActOnBehalfOfOtherIdentity"
                ])
            entries = self.engine.query(self.engine.ALL_DELEGATIONS_FILTER(), attributes)
        elif filter_ == "unconstrained":
            entries = self.engine.query(self.engine.UNCONSTRAINED_DELEGATION_FILTER(), attributes)
        elif filter_ == "constrained":
            if not verbose:
                attributes.append("msDS-AllowedToDelegateTo")
            entries = self.engine.query(self.engine.CONSTRAINED_DELEGATION_FILTER(), attributes)
        elif filter_ == "rbcd":
            if not verbose:
                attributes.append("msDS-AllowedToActOnBehalfOfOtherIdentity")
            entries = self.engine.query(self.engine.RESOURCE_BASED_CONSTRAINED_DELEGATION_FILTER(), attributes)
        else:
            return None

        if verbose:
            self.display(entries, verbose)
        else:
            for entry in entries:
                try:
                    uac = entry["userAccountControl"]
                    sam = entry["samAccountName"]
                    delegate = entry.get("msDS-AllowedToDelegateTo")
                    allowed_to_act = entry.get("msDS-AllowedToActOnBehalfOfOtherIdentity")
                    if (filter_ == "unconstrained" or filter_ == "all") and "TRUSTED_FOR_DELEGATION" in uac:
                        print(f"{sam}:unconstrained:")
                    if (filter_ == "constrained" or filter_ == "all") and delegate:
                        transition = "with" if "TRUSTED_TO_AUTH_FOR_DELEGATION" in uac else "without"
                        for a in delegate:
                            print(f"{sam}:constrained {transition} protocol transition:{a}")
                    if (filter_ == "rbcd" or filter_ == "all") and allowed_to_act:
                        sd = parse_ntSecurityDescriptor(allowed_to_act)
                        for ace in sd['DACL']['ACEs']:
                            try:
                                sid = ace.get('SID')
                                if not sid:
                                    continue
                                res = self.engine.resolve_sid(sid)
                                name = res[0]['samAccountName']
                                print(f"{name}:rbcd:{sam}")
                            except Exception:
                                print(f"{sid}:rbcd:{sam}")
                except Exception:
                    continue

    # GETTERS #

    def get_zone(self, kwargs):
        """
        Return the records of a DNS zone.

        Arguments:
            #dns_zone:string
                DNS zone to retrieve records
        """
        dns_zone = kwargs["dns_zone"]
        try:
            results = self.engine.query(
                self.engine.ZONE_FILTER(),
                base=','.join([f"DC={dns_zone}", "CN=MicrosoftDNS,DC=DomainDNSZones", self.engine.base_dn])
            )
        except LdapActiveDirectoryView.ActiveDirectoryLdapException as e:
            error(e)
        else:
            self.display(results)

    def get_membersof(self, kwargs):
        """
        List the members of `group`.

        Arguments:
            @verbose:bool
                Results will contain full information
            #group:string
                Group to list members
        """
        group = kwargs["group"]
        verbose = kwargs.get("verbose", False)

        results = self.engine.query(self.engine.GROUP_DN_FILTER(group), ["distinguishedName", "objectSid"])
        if results:
            group_dn = results[0]["distinguishedName"]
        else:
            error("Group {group} does not exists".format(group=group))

        primary_group_id = results[0]["objectSid"].split('-')[-1]
        results = self.engine.query(self.engine.ACCOUNTS_IN_GROUP_FILTER(primary_group_id, group_dn))
        self.display(results, verbose)

    def get_memberships(self, kwargs):
        """
        List the group for which `account` belongs to.

        Arguments:
            #account:string
                User to list memberships
            @recursive:bool
                List recursively the groups
        """
        account = kwargs["account"]
        recursive = kwargs.get("recursive", False)

        already_printed = set()

        def lookup_groups(dn, leading_sp, already_treated):
            results = self.engine.query(self.engine.DISTINGUISHED_NAME(dn), ["memberOf", "primaryGroupID"])
            for result in results:
                if "memberOf" in result:
                    for group_dn in result["memberOf"]:
                        if group_dn not in already_treated:
                            print("{g:>{width}}".format(g=group_dn, width=leading_sp + len(group_dn)))
                            already_treated.add(group_dn)
                            lookup_groups(group_dn, leading_sp + 4, already_treated)

                if "primaryGroupID" in result and result["primaryGroupID"]:
                    pid = result["primaryGroupID"]
                    results = self.engine.query(self.engine.PRIMARY_GROUP_ID(pid))
                    already_treated.add(results[0]["dn"])

            return already_treated

        results = self.engine.query(self.engine.ACCOUNT_IN_GROUPS_FILTER(account), ["memberOf", "primaryGroupID"])
        for result in results:
            if "memberOf" in result:
                for group_dn in result["memberOf"]:
                    print(group_dn)
                    if recursive:
                        already_printed.add(group_dn)
                        s = lookup_groups(group_dn, 4, already_printed)
                        already_printed.union(s)

            if "primaryGroupID" in result:
                pid = result["primaryGroupID"]
                results = self.engine.query(self.engine.PRIMARY_GROUP_ID(pid))
                if results:
                    print(results[0]["dn"])

    def get_from_sid(self, kwargs):
        """
        Return the object associated with the given `sid`.

        Arguments:
            @verbose:bool
                Results will contain full information
            #sid:string
                SID to search for
        """
        sid = kwargs["sid"]
        verbose = kwargs.get("verbose", False)

        try:
            result = self.engine.resolve_sid(sid)
            if isinstance(result, str):
                print(result)
            else:
                self.display(result, verbose)
        except ActiveDirectoryView.ActiveDirectoryInvalidSID:
            error("Invalid SID")

    def get_from_guid(self, kwargs):
        """
        Return the object associated with the given `guid`.

        Arguments:
            @verbose:bool
                Results will contain full information
            #guid:string
                GUID to search for
        """
        guid = kwargs["guid"]
        verbose = kwargs.get("verbose", False)

        try:
            self.display(self.engine.resolve_guid(guid), verbose)
        except ActiveDirectoryView.ActiveDirectoryLdapInvalidGUID:
            error("Invalid GUID")

    def get_object(self, kwargs):
        """
        Return the records containing `object` in a CN.

        Arguments:
            @verbose:bool
                Results will contain full information
            #object:string
                Pattern to look for in CNs
        """
        anr = kwargs["object"]
        verbose = kwargs.get("verbose", False)

        if verbose:
            attributes = ALL
        else:
            attributes = ["sAMAccountName", "objectClass"]
        results = self.engine.query(self.engine.ANR(anr), attributes)
        self.display(results, verbose)

    def get_sddl(self, kwargs):
        """
        Returns the SDDL of an object given it's CN.

        Arguments:
            #object:string
                CN of object.
        """
        anr = kwargs["object"]

        results = self.engine.get_sddl(f"(anr={anr})")

        self.display(results, True, False)

    def get_silo(self, kwargs):
        """
        Get information about a specific `silo`.

        Arguments:
            #silo:string
                Silo to query
            @information:string = ["all", "members", "auth_policies"]
                Information to query
        """
        silo = kwargs["silo"]
        info = kwargs["information"]

        if info == "all":
            attributes = ALL
        elif info == "members":
            attributes = ["msDS-AuthNPolicySiloMembers"]
        elif info == "auth_policies":
            attributes = ["msDS-ComputerAuthNPolicy", "msDS-ServiceAuthNPolicy", "msDS-UserAuthNPolicy"]
        else:
            return None

        results = self.engine.query(
            self.engine.SILO_FILTER(silo),
            attributes,
            base=','.join(["CN=AuthN Policy Configuration,CN=Services,CN=Configuration", self.engine.base_dn])
        )
        if not results:
            error(f"Silo {silo} does not exists")

        if info == "all":
            self.display(results, True)
        elif info == "members":
            members = results[0]["msDS-AuthNPolicySiloMembers"]
            print(*members, sep="\n")
        elif info == "auth_policies":
            print(f"msDS-ComputerAuthNPolicy: {results[0]['msDS-ComputerAuthNPolicy']}")
            print(f"msDS-ServiceAuthNPolicy: {results[0]['msDS-ServiceAuthNPolicy']}")
            print(f"msDS-UserAuthNPolicy: {results[0]['msDS-UserAuthNPolicy']}")

    # MISC #

    def misc_search(self, kwargs):
        """
        Query the LDAP with `filter` and retrieve ALL or `attributes` if specified.

        Arguments:
            #filter:string
                LDAP filter to search for
            #attributes:string = 'ALL'
                Comma separated list of attributes to display, ALL for every possible attribute
        """
        attr = kwargs["attributes"]
        filter_ = kwargs["filter"]

        try:
            if attr and attr != "ALL":
                results = self.engine.query(filter_, attr.split(","))
            else:
                results = self.engine.query(filter_)
            self.display(results, True)
        except PyAsn1UnicodeDecodeError as e:
            error("Decoding error with the filter")
        except Exception as e:
            if e.__str__() == "":
                error("An exception occurred with the provided filter")
            else:
                error(e)

    def misc_all(self, kwargs):
        """
        Collect and store computers, domain_policy, zones, gpo, groups, ou, users, trusts, pso information

        Arguments:
            #output:string
                File prefix for the files that will be created during the execution
        """
        output = kwargs["output"]
        kwargs["verbose"] = False

        for command, method in self.get_commands(prefix="list_"):
            info("Retrieving {command} output".format(command=command))
            if self.has_option(method, "filter"):
                filter_ = self.retrieve_default_val_for_arg(method, "filter")
                for f in filter_:
                    sys.stdout = Logger("{output}_{command}_{filter}.lst".format(output=output, command=command, filter=f), quiet=True)
                    kwargs["filter"] = f
                    getattr(self, method)(kwargs)

                    if self.has_option(method, "verbose"):
                        info("Retrieving {command} verbose output".format(command=command))
                        sys.stdout = Logger("{output}_{command}_{filter}.json".format(output=output, command=command, filter=f), quiet=True)
                        kwargs["verbose"] = True
                        getattr(self, method)(kwargs)
                        kwargs["verbose"] = False
                kwargs["filter"] = None
            else:
                sys.stdout = Logger("{output}_{command}.lst".format(output=output, command=command), quiet=True)
                getattr(self, method)(kwargs)

                if self.has_option(method, "verbose"):
                    info("Retrieving {command} verbose output".format(command=command))
                    sys.stdout = Logger("{output}_{command}.json".format(output=output, command=command), quiet=True)
                    kwargs["verbose"] = True
                    getattr(self, method)(kwargs)
                    kwargs["verbose"] = False

    def misc_enum_users(self, kwargs):
        """
        Anonymously enumerate users with LDAP pings.

        Arguments:
            #file:string
                File containing a list of usernames to try.
            @delay:int = 0
                Delay in milliseconds between each try.
        """

        # LDAP pings can only be used with an anonymous bind
        if self.engine.ldap.authentication != 'ANONYMOUS':
            error('The enum_users feature can only be used with an anonymous bind (-a option)')

        file = kwargs["file"]
        delay = kwargs["delay"]
        with open(file, 'r') as f:
            while True:
                l = f.readline()[:-1]
                if not l:
                    break
                if self.engine.user_exists(l):
                    print(l)
                sleep(delay / 1000)

    # ACTION #

    def action_unlock(self, kwargs):
        """
        Unlock `user`.

        Arguments:
            #user:string
                User to unlock
        """
        user = kwargs["user"]

        if self.engine.unlock(user):
            info("User {username} unlocked (or was already unlocked)".format(username=user))
        else:
            error("Unable to unlock {username}, check privileges".format(username=user))

    def action_modify_password(self, kwargs):
        """
        Change `user`'s password.

        Arguments:
            #user:string
                User to unlock
            #newpassword:string
                New password
            #currpassword:string = None
                Current password
        """
        user = kwargs["user"]
        new = kwargs["newpassword"]
        curr = kwargs.get("currpassword", None)
        if curr == "None":
            curr = None

        if self.engine.modify_password(user, curr, new):
            info("Password of {username} changed".format(username=user))
        else:
            error("Unable to change {username}'s password, check privileges or try with ldaps://".format(username=user))

    def action_add_to_group(self, kwargs):
        """
        Add `user` to `group`.

        Arguments:
            #user:string
                Target user (dn format). Ex: "CN=bob,CN=Users,DC=CORP,DC=LOCAL"
            #group:string
                Target group (dn format). Ex: "CN=Domain Admins,CN=Users,DC=CORP,DC=LOCAL"
        """
        user = kwargs["user"]
        group = kwargs["group"]

        if self.engine.add_user_to_group(user, group):
            info(f"User {user} sucessfully added to {group}")
        else:
            error(f"Unable to add {user} to {group}, check privileges or dn")

    def action_remove_from_group(self, kwargs):
        """
        Remove `user` from `group`.

        Arguments:
            #user:string
                Target user (dn format). Ex: "CN=bob,CN=Users,DC=CORP,DC=LOCAL"
            #group:string
                Target group (dn format). Ex: "CN=Domain Admins,CN=Users,DC=CORP,DC=LOCAL"
        """
        user = kwargs["user"]
        group = kwargs["group"]

        if self.engine.remove_user_from_group(user, group):
            info(f"User {user} sucessfully removed from {group}")
        else:
            error(f"Unable to remove {user} from {group}, check privileges or dn")


class MSDS_MANAGEDPASSWORD_BLOB(Structure):
    structure = (
        ('Version', '<H'),
        ('Reserved', '<H'),
        ('Length', '<L'),
        ('CurrentPasswordOffset', '<H'),
        ('PreviousPasswordOffset', '<H'),
        ('QueryPasswordIntervalOffset', '<H'),
        ('UnchangedPasswordIntervalOffset', '<H'),
        ('CurrentPassword', ':'),
        ('PreviousPassword', ':'),
        ('QueryPasswordInterval', ':'),
        ('UnchangedPasswordInterval', ':'),
    )

    def __init__(self, data=None):
        Structure.__init__(self, data=data)

    def fromString(self, data):
        Structure.fromString(self, data)

        if self['PreviousPasswordOffset'] == 0:
            endData = self['QueryPasswordIntervalOffset']
        else:
            endData = self['PreviousPasswordOffset']

        self['CurrentPassword'] = self.rawData[self['CurrentPasswordOffset']:][:endData - self['CurrentPasswordOffset']]
        if self['PreviousPasswordOffset'] != 0:
            self['PreviousPassword'] = self.rawData[self['PreviousPasswordOffset']:][:self['QueryPasswordIntervalOffset']-self['PreviousPasswordOffset']]

        self['QueryPasswordInterval'] = self.rawData[self['QueryPasswordIntervalOffset']:][:self['UnchangedPasswordIntervalOffset']-self['QueryPasswordIntervalOffset']]
        self['UnchangedPasswordInterval'] = self.rawData[self['UnchangedPasswordIntervalOffset']:]


def main():
    parser = ArgumentParser()
    parser.add_argument("-o", "--outfile", default="", help="Store the results in a file")
    parser.add_argument("--security_desc", action="store_true", help="Enable the retrieval of security descriptors in ldeep results")

    sub = parser.add_subparsers(title="Mode", dest="mode", description="Available modes", help="Backend engine to retrieve data")
    sub.required = True

    ldap = sub.add_parser("ldap", description="LDAP mode")
    cache = sub.add_parser("cache", description="Cache mode")

    ldap.add_argument("-d", "--domain", required=True, help="The domain as NetBIOS or FQDN")
    ldap.add_argument("-s", "--ldapserver", required=True, help="The LDAP path (ex : ldap://corp.contoso.com:389)")
    ldap.add_argument("-b", "--base", default="", help="LDAP base for query (by default, this value is pulled from remote Ldap)")
    ldap.add_argument("-t", "--type", default="ntlm", choices=['ntlm', 'simple'], help="Authentication type: ntlm (default) or simple")

    cache.add_argument("-d", "--dir", default=".", type=str, help="Use saved JSON files in specified directory as cache")
    cache.add_argument("-p", "--prefix", required=True, type=str, help="Prefix of ldeep saved files")

    ntlm = ldap.add_argument_group("NTLM authentication")
    ntlm.add_argument("-u", "--username", help="The username")
    ntlm.add_argument("-p", "--password", help="The password used for the authentication")
    ntlm.add_argument("-H", "--ntlm", help="NTLM hashes, format is LMHASH:NTHASH")

    kerberos = ldap.add_argument_group("Kerberos authentication")
    kerberos.add_argument("-k", "--kerberos", action="store_true", help="For Kerberos authentication, ticket file should be pointed by $KRB5NAME env variable")

    certificate = ldap.add_argument_group("Certificate authentication")
    certificate.add_argument("--pfx-file", help="PFX file")
    certificate.add_argument("--cert-pem", help="User certificate")
    certificate.add_argument("--key-pem", help="User private key")

    anonymous = ldap.add_argument_group("Anonymous authentication")
    anonymous.add_argument("-a", "--anonymous", action="store_true", help="Perform anonymous binds")

    Ldeep.add_subparsers(ldap, "ldap", ["list_", "get_", "misc_", "action_"], title="commands", description="available commands")
    Ldeep.add_subparsers(cache, "cache", ["list_", "get_"], title="commands", description="available commands")

    args = parser.parse_args()

    # Output
    if args.outfile:
        sys.stdout = Logger(args.outfile, quiet=False)

    cache = "prefix" in args  # figuring out whether we use the cache or not

    # main
    if cache:
        try:
            query_engine = CacheActiveDirectoryView(args.dir, args.prefix)
        except CacheActiveDirectoryView.CacheActiveDirectoryDirNotFoundException as e:
            error(e)

    else:
        try:
            # Authentication
            method = "NTLM"
            if args.kerberos:
                method = "Kerberos"
            elif args.cert_pem or args.pfx_file:
                method = "Certificate"
            elif args.anonymous or args.command_ldap == "enum_users":
                method = "anonymous"
            elif args.type == "ntlm":
                method = "NTLM"
            elif args.type == "simple":
                method = "SIMPLE"
            else:
                error("Lack of authentication options: either Kerberos, Certificate, Username with Password (can be a NTLM hash) or Anonymous.")

            query_engine = LdapActiveDirectoryView(args.ldapserver, args.domain, args.base, args.username, args.password, args.ntlm, args.pfx_file, args.cert_pem, args.key_pem, method)

        except LdapActiveDirectoryView.ActiveDirectoryLdapException as e:
            error(e)

    # If `security_desc` are requested, enable LDAP Security Descriptor flags and modify the default attributes
    # In cache mode, the security_desc corresponding JSON field will be kept
    if args.security_desc:
        query_engine.set_controls(LDAP_SERVER_SD_FLAGS_OID_SEC_DESC)
        query_engine.set_all_attributes([ALL_ATTRIBUTES, ALL_OPERATIONAL_ATTRIBUTES, "ntSecurityDescriptor"])

    ldeep = Ldeep(query_engine)

    try:
        ldeep.dispatch_command(args)
    except CacheActiveDirectoryView.CacheActiveDirectoryException as e:
        error(e)
    except NotImplementedError:
        error("Feature not yet available")


if __name__ == "__main__":
    main()



from sys import exit
from struct import unpack
from socket import inet_ntoa
from ssl import CERT_NONE

from ldap3 import Server, Connection, SASL, KERBEROS, NTLM, SUBTREE, ALL as LDAP3_ALL, BASE, DEREF_NEVER
from ldap3 import SIMPLE
from ldap3.protocol.formatters.formatters import format_sid
from ldap3.core.exceptions import LDAPOperationResult, LDAPSocketOpenError
from ldap3.extend.microsoft.unlockAccount import ad_unlock_account
from ldap3.extend.microsoft.modifyPassword import ad_modify_password
from ldap3.extend.microsoft.addMembersToGroups import ad_add_members_to_groups as addUsersInGroups
from ldap3.extend.microsoft.removeMembersFromGroups import ad_remove_members_from_groups as removeUsersInGroups

import ldap3

from ldeep.views.activedirectory import ActiveDirectoryView, ALL, validate_sid, validate_guid
from ldeep.views.constants import USER_ACCOUNT_CONTROL, DNS_TYPES, SAM_ACCOUNT_TYPE, PWD_PROPERTIES, TRUSTS_INFOS, WELL_KNOWN_SIDS, LOGON_SAM_LOGON_RESPONSE_EX
from ldeep.utils.sddl import parse_ntSecurityDescriptor

PAGE_SIZE = 1000


# define an ldap3-compliant formatters
def format_userAccountControl(raw_value):
    try:
        val = int(raw_value)
        result = []
        for k, v in USER_ACCOUNT_CONTROL.items():
            if v & val:
                result.append(k)
        return " | ".join(result)
    except (TypeError, ValueError):  # expected exceptions↲
        pass
    except Exception:  # any other exception should be investigated, anyway the formatters return the raw_value
        pass
    return raw_value


# define an ldap3-compliant formatters
def format_samAccountType(raw_value):
    try:
        val = int(raw_value)
        result = []
        for k, v in SAM_ACCOUNT_TYPE.items():
            if v & val:
                result.append(k)
        return " | ".join(result)
    except (TypeError, ValueError):  # expected exceptions↲
        pass
    except Exception:  # any other exception should be investigated, anyway the formatter returns the raw_value
        pass
    return raw_value


# define an ldap3-compliant formatters
def format_pwdProperties(raw_value):
    try:
        val = int(raw_value)
        result = []
        for k, v in PWD_PROPERTIES.items():
            if v & val:
                result.append(k)
        return " | ".join(result)
    except (TypeError, ValueError):  # expected exceptions↲
        pass
    except Exception:  # any other exception should be investigated, anyway the formatter returns the raw_value
        pass
    return raw_value

# define an ldap3-compliant formatters
def format_trustsInfos(raw_value):
    try:
        val = int(raw_value)
        result = []
        for k, v in TRUSTS_INFOS.items():
            if v & val:
                result.append(k)
        return " | ".join(result)
    except (TypeError, ValueError):  # expected exceptions↲
        pass
    except Exception:  # any other exception should be investigated, anyway the formatter returns the raw_value
        pass
    return raw_value


# define an ldap3-compliant formatters
def format_dnsrecord(raw_value):
    databytes = raw_value[0:4]
    datalen, datatype = unpack("HH", databytes)
    data = raw_value[24:24 + datalen]
    for recordname, recordvalue in DNS_TYPES.items():
        if recordvalue == datatype:
            if recordname == "A":
                target = inet_ntoa(data)
            else:
                # how, ugly
                data = data.decode('unicode-escape')
                target = ''.join([c for c in data if ord(c) > 31 or ord(c) == 9])
            return "%s %s" % (recordname, target)

def format_ad_timedelta(raw_value):
    """
    Convert a negative filetime value to an integer timedelta.
    """
    if isinstance(raw_value, bytes):
        raw_value = int(raw_value)
    return raw_value

def format_ad_timedelta(raw_value):
    """
    Convert a negative filetime value to an integer timedelta.
    """
    if isinstance(raw_value, bytes):
        raw_value = int(raw_value)
    return raw_value


#from http://www.kouti.com/tables/baseattributes.htm
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.8"] = (format_userAccountControl, None)
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.302"] = (format_samAccountType, None)
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.382"] = (format_dnsrecord, None)
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.121"] = (format_sid, None)
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.93"] = (format_pwdProperties, None)
ldap3.protocol.formatters.standard.standard_formatter["1.2.840.113556.1.4.382"] = (format_dnsrecord, None)
ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.60'] = (format_ad_timedelta, None)
ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.74'] = (format_ad_timedelta, None)
ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.78'] = (format_ad_timedelta, None)
ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.4.470'] = (format_trustsInfos, None)
ldap3.protocol.formatters.standard.standard_formatter['1.2.840.113556.1.2.281'] = (parse_ntSecurityDescriptor, None)

class LdapActiveDirectoryView(ActiveDirectoryView):
    """
    Manage a LDAP connection to a LDAP Active Directory.
    """

    # Constant functions
    USER_LOCKED_FILTER = lambda _: "(&(objectCategory=Person)(objectClass=user)(lockoutTime:1.2.840.113556.1.4.804:=4294967295))"
    GROUPS_FILTER = lambda _: "(objectClass=group)"
    ZONES_FILTER = lambda _: "(&(objectClass=dnsZone)(!(dc=RootDNSServers)))"
    ZONE_FILTER = lambda _: "(objectClass=dnsNode)"
    SITES_FILTER = lambda _: "(objectClass=site)"
    SUBNET_FILTER = lambda _, s: f"(SiteObject={s})"
    PKI_FILTER = lambda _: "(objectclass=pKIEnrollmentService)"
    USER_ALL_FILTER = lambda _: "(&(objectCategory=Person)(objectClass=user))"
    USER_SPN_FILTER = lambda _: "(&(objectCategory=Person)(objectClass=user)(servicePrincipalName=*)(!(sAMAccountName=krbtgt)))"
    USER_ACCOUNT_CONTROL_FILTER = lambda _, n: f"(&(objectCategory=Person)(objectClass=user)(userAccountControl:1.2.840.113556.1.4.803:={n}))"
    USER_ACCOUNT_CONTROL_FILTER_NEG = lambda _, n: f"(&(objectCategory=Person)(objectClass=user)(!(userAccountControl:1.2.840.113556.1.4.803:={n})))"
    ANR = lambda _, u: f"(anr={u})"
    DISTINGUISHED_NAME = lambda _, n: f"(distinguishedName={n})"
    COMPUTERS_FILTER = lambda _: "(objectClass=computer)"
    GROUP_DN_FILTER = lambda _, g: f"(&(objectClass=group)(sAMAccountName={g}))"
    USER_DN_FILTER = lambda _, u: f"(&(objectClass=user)(objectCategory=Person)(sAMAccountName={u}))"
    ACCOUNTS_IN_GROUP_FILTER = lambda _, p, g: f"(|(primaryGroupID={p})(memberOf={g}))"
    ACCOUNT_IN_GROUPS_FILTER = lambda _, u: f"(sAMAccountName={u})"
    PRIMARY_GROUP_ID = lambda s, i: f"(objectSid={s.get_domain_sid()}-{i})"
    DOMAIN_INFO_FILTER = lambda _: "(objectClass=domain)"
    GPO_INFO_FILTER = lambda _: "(objectCategory=groupPolicyContainer)"
    PSO_INFO_FILTER = lambda _: "(objectClass=msDS-PasswordSettings)"
    TRUSTS_INFO_FILTER = lambda _: "(objectCategory=trustedDomain)"
    OU_FILTER = lambda _: "(|(objectClass=OrganizationalUnit)(objectClass=domain))"
    ENUM_USER_FILTER = lambda _, n: f"(&(NtVer=\x06\x00\x00\x00)(AAC=\x10\x00\x00\x00)(User={n}))"
    ALL_FILTER = lambda _: "(objectClass=*)"
    AUTH_POLICIES_FILTER = lambda _: "(objectClass=msDS-AuthNPolicy)"
    SILOS_FILTER = lambda _: "(objectClass=msDS-AuthNPolicySilo)"
    SILO_FILTER = lambda _, s: f"(&(objectClass=msDS-AuthNPolicySilo)(cn={s}))"
    SMSA_FILTER = lambda _: "(ObjectClass=msDS-ManagedServiceAccount)"
    SHADOW_PRINCIPALS_FILTER = lambda _: "(objectclass=msDS-ShadowPrincipal)"
    UNCONSTRAINED_DELEGATION_FILTER = lambda _: f"(userAccountControl:1.2.840.113556.1.4.803:=524288)"
    CONSTRAINED_DELEGATION_FILTER = lambda _: f"(msDS-AllowedToDelegateTo=*)"
    RESOURCE_BASED_CONSTRAINED_DELEGATION_FILTER = lambda _: f"(msDS-AllowedToActOnBehalfOfOtherIdentity=*)"
    ALL_DELEGATIONS_FILTER = lambda _: f"(|(userAccountControl:1.2.840.113556.1.4.803:=524288)(msDS-AllowedToDelegateTo=*)(msDS-AllowedToActOnBehalfOfOtherIdentity=*))"

    class ActiveDirectoryLdapException(Exception):
        pass

    def __init__(self, server, domain="", base="", username="", password="", ntlm="", pfx_file="", cert_pem="", key_pem="", method="NTLM"):
        """
        LdapActiveDirectoryView constructor.
        Initialize the connection with the LDAP server.

        Three authentication modes:
            * Kerberos (ldap3 will automatically retrieve the $KRB5CCNAME env variable)
            * NTLM (username + NTLM hash/password)
            * SIMPLE (username + password)

        @server: Server to connect and perform LDAP query to.
        @domain: Fully qualified domain name of the Active Directory domain.
        @base: Base for the LDAP queries.
        @username: Username to use for the authentication
        @password: Password to use for the authentication (for SIMPLE authentication)
        @ntlm: NTLM hash to use for the authentication (for NTLM authentication)
        @method: Either to use NTLM, SIMPLE, Kerberos or anonymous authentication.

        @throw ActiveDirectoryLdapException when the connection or the bind does not work.
        """
        self.username = username
        self.password = password
        self.ntlm = ntlm
        self.pfx_file = pfx_file
        self.cert = cert_pem
        self.key = key_pem
        self.server = server
        self.domain = domain
        self.hostnames = []

        self.set_controls()
        self.set_all_attributes()

        if method == "Certificate":
            if not self.server.startswith("ldaps"):
                # TODO start tls if ldap
                print("At this moment ldeep needs to use ldaps (use ldaps:// before the server parameter)")
                exit(1)
            else:
                if self.pfx_file:
                    # TODO deal with pfx password protected
                    with open(pfx_file, 'rb') as f:
                        pfxdata = f.read()
                    from cryptography.hazmat.primitives.serialization import pkcs12
                    from cryptography.hazmat.primitives import serialization
                    privkey, cert, extra_certs = pkcs12.load_key_and_certificates(pfxdata, None)
                    key = privkey.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.TraditionalOpenSSL,
                        encryption_algorithm=serialization.NoEncryption(),
                    )
                    cert = cert.public_bytes(encoding=serialization.Encoding.PEM)
                    try:
                        from tempfile import gettempdir
                        key_path = f'{gettempdir()}/ldeep_key'
                        cert_path = f'{gettempdir()}/ldeep_cert'
                        with open(key_path, "wb") as f1, open(cert_path, "wb") as f2:
                            f1.write(key)
                            f2.write(cert)
                        tls = ldap3.Tls(local_private_key_file=key_path, local_certificate_file=cert_path, validate=CERT_NONE)
                    except PermissionError:
                        print("Can't write key and cert to disk")
                        exit(1)
                else:
                    tls = ldap3.Tls(local_private_key_file=self.key, local_certificate_file=self.cert, validate=CERT_NONE)
        else:
            tls=ldap3.Tls(validate=CERT_NONE)

        if self.server.startswith("ldaps"):
            server = Server(
                self.server,
                port=636,
                use_ssl=True,
                allowed_referral_hosts=[('*', True)],
                get_info=LDAP3_ALL,
                tls=tls
            )
        else:
            server = Server(self.server, get_info=LDAP3_ALL)

        if method == "Kerberos":
            self.ldap = Connection(server, authentication=SASL, sasl_mechanism=KERBEROS)
        elif method == "Certificate":
            self.ldap = Connection(server)
        elif method == "anonymous":
            self.ldap = Connection(server)
        elif method == "NTLM":
            if password is not None:
                ntlm = password
            else:
                try:
                    lm, nt = ntlm.split(':')
                    lm = "aad3b435b51404eeaad3b435b51404ee" if not lm else lm
                    ntlm = f"{lm}:{nt}"
                except Exception as e:
                    print(e)
                    print("Incorrect hash, format is LMHASH:NTHASH")
                    exit(1)
            self.ldap = Connection(
                server,
                user=f"{domain}\\{username}",
                password=ntlm,
                authentication=NTLM, check_names=True
            )
        elif method == "SIMPLE":
            if not password:
                print("Password is required (-p)")
                exit(1)
            if "." in domain:
                domain, _, _ = domain.partition(".")
            self.ldap = Connection(
                server,
                user=f"{domain}\\{username}",
                password=password,
                authentication=SIMPLE, check_names=True
            )

        try:
            if method == "Certificate":
                import os
                try:
                    self.ldap.open()
                    if self.pfx_file:
                        os.remove(key_path)
                        os.remove(cert_path)
                except ldap3.core.exceptions.LDAPSocketOpenError:
                    print("Cannot get private key data, corrupted key or wrong passphrase ?")
                    if self.pfx_file:
                        os.remove(key_path)
                        os.remove(cert_path)
                    exit(1)
                except Exception as e:
                    print("Unhandled Exception")
                    import traceback
                    traceback.print_exc()
                    exit(1)
            else:
                if not self.ldap.bind():
                    raise self.ActiveDirectoryLdapException("Unable to bind with provided information")
        except LDAPSocketOpenError:
            raise self.ActiveDirectoryLdapException(f"Unable to open connection with {self.server}")

        self.base_dn = base or server.info.other["defaultNamingContext"][0]
        self.fqdn = ".".join(map(lambda x: x.replace("DC=", ''), self.base_dn.split(',')))
        self.search_scope = SUBTREE

    def set_controls(self, controls=[]):
        self.controls = controls

    def set_all_attributes(self, attributes=ALL):
        self.attributes = attributes

    def all_attributes(self):
        return self.attributes

    def query(self, ldapfilter, attributes=[], base=None, scope=None):
        """
        Perform a query to the LDAP server and return the results.

        @ldapfilter: The LDAP filter to query (see RFC 2254).
        @attributes: List of attributes to retrieved with the query.
        @base: Base to use during the request.
        @scope: Scope to use during the request.

        @return a list of records.
        """
        attributes = self.attributes if attributes == [] else attributes
        result_set = []
        try:
            entry_generator = self.ldap.extend.standard.paged_search(
                search_base=base or self.base_dn,
                search_filter=ldapfilter,
                search_scope=scope or self.search_scope,
                attributes=attributes,
                controls=self.controls,
                paged_size=PAGE_SIZE,
                generator=True
            )

            for entry in entry_generator:
                if "dn" in entry:
                    d = entry["attributes"]
                    d["dn"] = entry["dn"]
                    result_set.append(d)

        except LDAPOperationResult as e:
            raise self.ActiveDirectoryLdapException(e)

        return result_set

    def get_domain_sid(self):
        """
        Return the current domain SID by issueing a LDAP request.

        @return the domain sid or None if a problem occurred.
        """
        results = self.query(self.DOMAIN_INFO_FILTER(), ["ObjectSid"])

        if results:
            return results[0]["objectSid"]

        return None


    def resolve_sid(self, sid):
        """
        Two cases:
            * the SID is a WELL KNOWN SID and a local SID, the name of the corresponding account is returned;
            * else, the SID is search through the LDAP and the corresponding record is returned.

        @sid: the sid to search for.

        @throw ActiveDirectoryInvalidSID if the SID is not a valid SID.
        @return the record corresponding to the SID queried.
        """
        if sid in WELL_KNOWN_SIDS:
            return WELL_KNOWN_SIDS[sid]
        elif validate_sid(sid):
            results = self.query(f"(&(ObjectSid={sid}))")
            if results:
                return results
        raise self.ActiveDirectoryInvalidSID(f"SID: {sid}")

    def resolve_guid(self, guid):
        """
        Return the LDAP record with the provided GUID.

        @guid: the guid to search for.

        @throw ActiveDirectoryInvalidGUID if the GUID is not a valid GUID.
        @return the record corresponding to the guid queried.
        """
        if validate_guid(guid):
            results = self.query(f"(&(ObjectGUID={guid}))")
            # Normally only one result should have been retrieved:
            if results:
                return results
        raise self.ActiveDirectoryInvalidGUID(f"GUID: {guid}")

    def get_sddl(self, ldapfilter, base=None, scope=None):
        """
        Perform a query to the LDAP server and return the results.

        @ldapfiler: The LDAP filter to query (see RFC 2254).
        @attributes: List of attributes to retrieved with the query.
        @base: Base to use during the request.
        @scope: Scope to use during the request.

        @return a list of records.
        """
        result_set = []
        try:
            result = self.ldap.search(
                search_base=base or self.base_dn,
                search_filter=ldapfilter,
                search_scope=scope or self.search_scope,
                attributes=['ntSecurityDescriptor'],
                controls=[('1.2.840.113556.1.4.801', True, b'\x30\x03\x02\x01\x07')]
            )

            if not result:
                raise self.ActiveDirectoryLdapException()
            else:
                for entry in self.ldap.response:
                    if "dn" in entry:
                        d = entry["attributes"]
                        d["dn"] = entry["dn"]
                        result_set.append(d)

        except LDAPOperationResult as e:
            raise self.ActiveDirectoryLdapException(e)

        return result_set

    def unlock(self, username):
        """
        Unlock an account.

        @username: the username associated to the account to unlock.

        @throw ActiveDirectoryLdapException if the account does not exist or the query returns more than one result.
        @return True if the account was successfully unlock or False otherwise.
        """
        results = self.query(self.USER_DN_FILTER(username))
        if len(results) != 1:
            raise self.ActiveDirectoryLdapException("Zero or non uniq result")
        else:
            user = results[0]
            unlock = ad_unlock_account(self.ldap, user["dn"])
            # goddamn, return value is either True or str...
            return isinstance(unlock, bool)

    def modify_password(self, username, oldpassword, newpassword):
        """
        Change the password of `username`.

        @username: the username associated to the account to modify its password.
        @newpassword: the new password to apply.
        @oldpassword: the old password.

        @throw ActiveDirectoryLdapException if the account does not exist or the query returns more than one result.
        @return True if the account was successfully unlock or False otherwise.
        """
        results = self.query(self.USER_DN_FILTER(username))
        if len(results) != 1:
            raise self.ActiveDirectoryLdapException("Zero or non uniq result")
        else:
            user = results[0]
            return ad_modify_password(self.ldap, user["dn"], newpassword, None)

    def add_user_to_group(self, user_dn, group_dn):
        """
        Add user to a group.

        @username: the username that will be added to the group. DN format: "CN=username,CN=Users,DC=CORP,DC=LOCAL"
        @group: the target group. DN format: "CN=group,CN=Users,DC=CORP,DC=LOCAL"

        @return True if the account was successfully added or False otherwise.
        """
        try:
            return addUsersInGroups(self.ldap, user_dn, group_dn)
        except ldap3.core.exceptions.LDAPInvalidDnError as e:
            # catch invalid group dn
            return False

    def remove_user_from_group(self, user_dn, group_dn):
        """
        Remove user from a group.

        @username: the username that will be removed from the group. dn format: "CN=username,CN=Users,DC=CORP,DC=LOCAL"
        @group: the target group. dn format: "CN=group,CN=Users,DC=CORP,DC=LOCAL"

        @return True if the account was successfully removed or if the account doesn't exist or False otherwise.
        """
        try:
            return removeUsersInGroups(self.ldap, user_dn, group_dn, fix=True)
        except ldap3.core.exceptions.LDAPInvalidDnError as e:
            # catch invalid group dn
            return False

    def user_exists(self, username):
        """
        Perform an LDAP ping to determine if the specified user exists.

        @username: the username to test.

        @return True if the user exists, False otherwise.
        """
        try:
            result = self.ldap.search(
                '',
                search_filter=self.ENUM_USER_FILTER(username),
                search_scope=BASE,
                attributes=['NetLogon'],
                dereference_aliases=DEREF_NEVER
            )

            if not result:
                raise self.ActiveDirectoryLdapException()
            else:
                for entry in self.ldap.response:
                    attr = entry.get('raw_attributes')
                    if attr:
                        netlogon = attr.get('netlogon')
                        if netlogon and len(netlogon[0]) > 1 and netlogon[0][:2] == LOGON_SAM_LOGON_RESPONSE_EX:
                            return True

        except LDAPOperationResult as e:
            raise self.ActiveDirectoryLdapException(e)

        return False




DNS_TYPES = {
	"ZERO": 0x0000,
	"A": 0x0001,
	"NS": 0x0002,
	"MD": 0x0003,
	"MF": 0x0004,
	"CNAME": 0x0005,
	"SOA": 0x0006,
	"MB": 0x0007,
	"MG": 0x0008,
	"MR": 0x0009,
	"NULL": 0x000A,
	"WKS": 0x000B,
	"PTR": 0x000C,
	"HINFO": 0x000D,
	"MINFO": 0x000E,
	"MX": 0x000F,
	"TXT": 0x0010,
	"RP": 0x0011,
	"AFSDB": 0x0012,
	"X25": 0x0013,
	"ISDN": 0x0014,
	"RT": 0x0015,
	"SIG": 0x0018,
	"KEY": 0x0019,
	"AAAA": 0x001C,
	"LOC": 0x001D,
	"NXT": 0x001E,
	"SRV": 0x0021,
	"ATMA": 0x0022,
	"NAPTR": 0x0023,
	"DNAME": 0x0027,
	"DS": 0x002B,
	"RRSIG": 0x002E,
	"NSEC": 0x002F,
	"DNSKEY": 0x0030,
	"DHCID": 0x0031,
	"NSEC3": 0x0032,
	"NSEC3PARAM": 0x0033,
	"TLSA": 0x0034,
	"ALL": 0x00FF,
	"WINS": 0xFF01,
	"WINSR": 0xFF02,
}

USER_ACCOUNT_CONTROL = {
	"SCRIPT": 0x0001,
	"ACCOUNTDISABLE": 0x0002,
	"HOMEDIR_REQUIRED": 0x0008,
	"LOCKOUT": 0x0010,
	"PASSWD_NOTREQD": 0x0020,
	"PASSWD_CANT_CHANGE": 0x0040,
	"ENCRYPTED_TEXT_PWD_ALLOWED": 0x0080,
	"TEMP_DUPLICATE_ACCOUNT": 0x0100,
	"NORMAL_ACCOUNT": 0x0200,
	"INTERDOMAIN_TRUST_ACCOUNT": 0x0800,
	"WORKSTATION_TRUST_ACCOUNT": 0x1000,
	"SERVER_TRUST_ACCOUNT": 0x2000,
	"DONT_EXPIRE_PASSWORD": 0x10000,
	"MNS_LOGON_ACCOUNT": 0x20000,
	"SMARTCARD_REQUIRED": 0x40000,
	"TRUSTED_FOR_DELEGATION": 0x80000,
	"NOT_DELEGATED": 0x100000,
	"USE_DES_KEY_ONLY": 0x200000,
	"DONT_REQ_PREAUTH": 0x400000,
	"PASSWORD_EXPIRED": 0x800000,
	"TRUSTED_TO_AUTH_FOR_DELEGATION": 0x1000000,
	"PARTIAL_SECRETS_ACCOUNT": 0x04000000,
}

SAM_ACCOUNT_TYPE = {
	"SAM_DOMAIN_OBJECT": 0x0,
	"SAM_GROUP_OBJECT": 0x10000000,
	"SAM_NON_SECURITY_GROUP_OBJECT": 0x10000001,
	"SAM_ALIAS_OBJECT": 0x20000000,
	"SAM_NON_SECURITY_ALIAS_OBJECT": 0x20000001,
	"SAM_USER_OBJECT": 0x30000000,
	"SAM_NORMAL_USER_ACCOUNT": 0x30000000,
	"SAM_MACHINE_ACCOUNT": 0x30000001,
	"SAM_TRUST_ACCOUNT": 0x30000002,
	"SAM_APP_BASIC_GROUP": 0x40000000,
	"SAM_APP_QUERY_GROUP": 0x40000001,
	"SAM_ACCOUNT_TYPE_MAX": 0x7fffffff,
}

PWD_PROPERTIES = {
	"DOMAIN_PASSWORD_COMPLEX": 0x1,
	"DOMAIN_PASSWORD_NO_ANON_CHANGE": 0x2,
	"DOMAIN_PASSWORD_NO_CLEAR_CHANGE": 0x4,
	"DOMAIN_LOCKOUT_ADMINS": 0x8,
	"DOMAIN_PASSWORD_STORE_CLEARTEXT": 0x10,
	"DOMAIN_REFUSE_PASSWORD_CHANGE": 0x20
}

TRUSTS_INFOS = {
	"NON_TRANSITIVE": 0x1,
	"UPLEVEL_ONLY": 0x2,
	"QUARANTINED_DOMAIN": 0x4,
	"FOREST_TRANSITIVE": 0x8,
	"CROSS_ORGANIZATION": 0x10,
	"WITHIN_FOREST": 0x20,
	"TREAT_AS_EXTERNAL": 0x40,
	"USES_RC4_ENCRYPTION": 0x80,
	"USES_AES_KEYS": 0x100,
	"CROSS_ORGANIZATION_NO_TGT_DELEGATION": 0x200,
	"PIM_TRUST": 0x400
}

WELL_KNOWN_SIDS = {
	"S-1-5-32-544"	: r"BUILTIN\Administrators",
	"S-1-5-32-545"	: r"BUILTIN\Users",
	"S-1-5-32-546"	: r"BUILTIN\Guests",
	"S-1-5-32-547"	: r"BUILTIN\Power Users",
	"S-1-5-32-548"	: r"BUILTIN\Account Operators",
	"S-1-5-32-549"	: r"BUILTIN\Server Operators",
	"S-1-5-32-550"	: r"BUILTIN\Print Operators",
	"S-1-5-32-551"	: r"BUILTIN\Backup Operators",
	"S-1-5-32-552"	: r"BUILTIN\Replicators",
	"S-1-5-64-10"	: r"BUILTIN\NTLM Authentication",
	"S-1-5-64-14"	: r"BUILTIN\SChannel Authentication",
	"S-1-5-64-21"	: r"BUILTIN\Digest Authentication",
	"S-1-16-4096"	: r"BUILTIN\Low Mandatory Level",
	"S-1-16-8192"	: r"BUILTIN\Medium Mandatory Level",
	"S-1-16-8448"	: r"BUILTIN\Medium Plus Mandatory Level",
	"S-1-16-12288"	: r"BUILTIN\High Mandatory Level",
	"S-1-16-16384"	: r"BUILTIN\System Mandatory Level",
	"S-1-16-20480"	: r"BUILTIN\Protected Process Mandatory Level",
	"S-1-16-28672"	: r"BUILTIN\Secure Process Mandatory Level",
	"S-1-5-32-554"	: r"BUILTIN\Pre-Windows 2000 Compatible Access",
	"S-1-5-32-555"	: r"BUILTIN\Remote Desktop Users",
	"S-1-5-32-556"	: r"BUILTIN\Network Configuration Operators",
	"S-1-5-32-557"	: r"BUILTIN\Incoming Forest Trust Builders",
	"S-1-5-32-558"	: r"BUILTIN\Performance Monitor Users",
	"S-1-5-32-559"	: r"BUILTIN\Performance Log Users",
	"S-1-5-32-560"	: r"BUILTIN\Windows Authorization Access Group",
	"S-1-5-32-561"	: r"BUILTIN\Terminal Server License Servers",
	"S-1-5-32-562"	: r"BUILTIN\Distributed COM Users",
	"S-1-5-32-569"	: r"BUILTIN\Cryptographic Operators",
	"S-1-5-32-573"	: r"BUILTIN\Event Log Readers",
	"S-1-5-32-574"	: r"BUILTIN\Certificate Service DCOM Access",
	"S-1-5-32-575"	: r"BUILTIN\RDS Remote Access Servers",
	"S-1-5-32-576"	: r"BUILTIN\RDS Endpoint Servers",
	"S-1-5-32-577"	: r"BUILTIN\RDS Management Servers",
	"S-1-5-32-578"	: r"BUILTIN\Hyper-V Administrators",
	"S-1-5-32-579"	: r"BUILTIN\Access Control Assistance Operators",
	"S-1-5-32-580"	: r"BUILTIN\Remote Management Users",
}

FILETIME_FIELDS = [
	"badPasswordTime",
	"lastLogon",
	"lastLogoff",
	"lastLogonTimestamp",
	"pwdLastSet",
	"accountExpires",
	"lockoutTime"
]

DATETIME_FIELDS = [
	"dSCorePropagationData",
	"whenChanged",
	"whenCreated"
]

LDAP_SERVER_SD_FLAGS_OID_SEC_DESC = [('1.2.840.113556.1.4.801', True, b'\x30\x03\x02\x01\x07')]

LOGON_SAM_LOGON_RESPONSE_EX = b'\x17\x00'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import codecs
from setuptools import setup, find_packages
import sys

dirname = os.path.dirname(__file__)
version = open(os.path.join(dirname, "VERSION")).read().strip()
requirements = [line[:-1] for line in open(os.path.join(dirname, "requirements.txt"), "r").readlines()]

setup(
	# Basic info
	name="ldeep",
	version=version,
	author="b0z, flgy",
	author_email="bastien@faure.io, florian.guilbert@synacktiv.com",
	keywords='pentesting security windows active-directory networks',
	url="https://github.com/franc-pentest/ldeep",
	license="MIT",
	description="In-depth ldap enumeration utility",
	long_description=codecs.open("README.rst", "rb", "utf8").read(),

	# Classifiers (see https://pypi.python.org/pypi?%3Aaction=list_classifiers)
	classifiers=[
		"Development Status :: 4 - Beta",
		"Intended Audience :: Information Technology",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3.3",
		"Programming Language :: Python :: 3 :: Only",
		"Topic :: Security",
		"Operating System :: OS Independent",
	],

	# Packages and dependencies
	#package_dir={"": "ldeep"},
	packages=find_packages(include=["ldeep", "ldeep.*"]),
	install_requires=requirements,

	# Other configurations
	zip_safe=True,
	platforms='any',
	# entry points
	entry_points={
		"console_scripts": [
			"ldeep = ldeep.__main__:main"
		]
	},
	python_requires=">=3.3"
)
